/* Distributed under the terms of GPLv3 or later. See COPYING for details. */

/*! \file types.h
 *
 * \brief Definition of fundamental types in use.
 */

#ifndef INCLUDE_TYPES_H_
#define INCLUDE_TYPES_H_

#include <complex.h>

typedef __complex__ double dcomplex;

#ifdef USE_LAPACK
#ifdef USE_MKL
#ifndef MKL_INT 
#define MKL_INT int64_t
#endif
#include <mkl_lapacke.h>
#else
#include <lapacke.h>
#endif
#endif

#ifndef np_int
#ifdef lapack_int
#define np_int lapack_int
#else
#define np_int int64_t
#endif
#endif

/*! \brief Get the real part of a complex number.
 *
 * \param z: `complex double` The argument of the function.
 * \return rz: `double` The real part of the argument.
 */
double real(dcomplex z);

/*! \brief Get the imaginary part of a complex number.
 *
 * \param z: `complex double` The argument of the function.
 * \return iz: `double` The imaginary part of the argument.
 */
double imag(dcomplex z);
#endif
